"""
Core methods used for evaluating BPA and Feild logs,
by Luis A. Leiva and Jeff Huang.
Dual licensed under the MIT and GPL licenses.
"""

from __future__ import division
import os, sys, math
from scripts import mathlib
from scripts.utils import *
from scripts.parser import EvLogParser, ToolbarLogParser, EvXMLParser

def evalCompression(originalPoints, compressedPoints, clicktimes):
  dd = evalAtDistance(originalPoints, compressedPoints)
  dc = evalAtClicks(originalPoints, compressedPoints, clicktimes) if clicktimes else None
  return dd, dc

def evalAtDistance(originalPoints, compressedPoints):
  distances = []
  compressedTimes = [tc for tc,xc,yc in compressedPoints]
  for to,xo,yo in originalPoints:
    (xc,yc) = getPointBefore(compressedPoints, to)
    op, cp = [xo, yo], [xc, yc]
    distances.append( math.sqrt(mathlib.sqL2(op,cp)) )
  return sum(distances) / len(distances)
  
def evalAtClicks(originalPoints, compressedPoints, clicktimes):
  distances = []
  for ct in clicktimes:
    op = getPointBefore(originalPoints, ct)
    cp = getPointBefore(compressedPoints, ct)
    distances.append( math.sqrt(mathlib.sqL2(op,cp)) )
  return sum(distances) / len(distances)

def getPointBefore(points, timestamp):
  tf,xf,yf = points[0]
  tl,xl,yl = points[-1]
  prevCp = (xf,yf)
  for t,x,y in points:
    if timestamp <= t:
      return prevCp
    prevCp = (x,y)
  return (xl,yl)

def getPointAfter(points, timestamp):
  tl,xl,yl = points[-1]
  for t,x,y in points:
    if t >= timestamp:
      return (x,y)
  return (xl,yl)
      
def reformatPoints(points):
  return map(lambda (x,y,t): (t,x,y), [tup[1:4] for tup in points])

def assignTimestamps(points, timestamps):
  assert len(points) == len(timestamps)
  return [ (timestamps[i], pt[0], pt[1]) for i, pt in enumerate(points) ]

def cleanCoordsByIndex(track, indices):
  return map(lambda (x,y): [x,y], [track[i][1:3] for i in indices])

def getLogs():
  files = sys.stdin.readlines()
  logs = []
  for filename in files:
    filename = filename.strip()
    basename, extension = os.path.basename(filename).split('.')
    if extension == 'log':
      log = EvLogParser(filename)
    elif extension == 'txt':
      log = ToolbarLogParser(filename)
    elif extension == 'xml':
      log = EvXMLParser(filename)
    else:
      sys.exit("Error: Unknown log extension")
    logs.append(log)
  return logs
