"""
This script applies lossless compression over a mouse cursor dataset.
A `lossless.res' file will be created on the script dir.
"""

# Usage example (LAB data):  find ./logs-feild/ -name "toolbar_log.txt" -print | python eval-lossless.py
# Usage example (LIVE data): ls ./logs-bpa/*.log | python eval-lossless.py

from __future__ import division
import sys, math, os, time
from core import *
from scripts import mathlib
from scripts.utils import *
from scripts.delta import deltaCompression
from scripts.huffman import Huffman
from scripts.lzw import LZW
import zlib
import lz4 # Get it at http://pypi.python.org/pypi/lz4


def processCursor(log):
  result = {}
  track  = log.format(log.grep('move'))
  coords = map(lambda (x,y,t): [x,y,t], [tup[1:4] for tup in track])
  coordstr = str(coords)
  siz_orig = len(coordstr)
  
  # Method 1: Huffman
  huf = Huffman()
  start_time = time.time()
  compressed = huf.encode(coordstr)
  compress_time = time.time() - start_time
  original = huf.decode(compressed)
  assert coordstr == original
  siz_comp = (len(compressed)+7)/8 # convert bits to bytes
  siz_tree = len(str(huf.tree))
  fillResults(result, 'Huffman', compress_time, siz_comp + siz_tree, siz_orig)

  # Method 2: LZW
  lzw = LZW()
  start_time = time.time()
  compressed = lzw.encode(coordstr)
  compress_time = time.time() - start_time
  original = lzw.decode(compressed)
  assert coordstr == original
  siz_comp = len(compressed)
  fillResults(result, 'LZW', compress_time, siz_comp, siz_orig)
  
  # Method 3: Gzip
  start_time = time.time()
  compressed = coordstr.encode('zlib')
  compress_time = time.time() - start_time
  original = compressed.decode('zlib')
  assert coordstr == original
  siz_comp = len(compressed)  
  fillResults(result, 'Gzip', compress_time, siz_comp, siz_orig)
  
  # Method 4: LZ4
  start_time = time.time()
  compressed = lz4.dumps(coordstr)
  compress_time = time.time() - start_time
  original = lz4.loads(compressed)
  assert coordstr == original
  siz_comp = len(compressed)
  fillResults(result, 'LZ4', compress_time, siz_comp, siz_orig)

  # Method 5: Delta
  start_time = time.time()
  compressed = deltaCompression(coords)
  compress_time = time.time() - start_time
  siz_comp = len(str(compressed))
  fillResults(result, 'Delta', compress_time, siz_comp, siz_orig)  
  
  # That's all folks!
  return result

def fillResults(result, technique, compress_time, siz_comp, siz_orig):
  result[technique] = {
    "timesComp": compress_time * 1000,          # ms, lower is better
    "compRatio": (1 - siz_comp/siz_orig) * 100, # %, higher is better
    "bytesComp": siz_comp,                      # compressed size
    "bytesOrig": siz_orig,                      # original size
  }
  
def reportStruct():
  report = {}
  values = ["timesComp", "compRatio", "bytesComp", "bytesOrig"]
  technq = ["Hufman", "LZW", "Gzip", "LZ4", "Delta"]
  for t in technq:
    if t not in report:
      report[t] = {}
    for v in values:
      if v not in report[t]:
        report[t][v] = []
  return report
  
def main():
  final_report = reportStruct()
  browse_times = []
  for num,log in enumerate(getLogs()): # reads from stdin
    res = processCursor(log)
    for technique,results in res.iteritems():
      for metric,value in sorted(results.iteritems()):
        final_report[technique][metric].append(value)
    browse_times.append(log.sesstime()/60) # minutes
    if num > 0 and num % 1000 == 0:
      print num,"logs processed"
  print
  m, sd = mathlib.msd(browse_times)
  n = len(browse_times) # We need this to build CIs offline
  print "%d users browsed for %f minutes (SD=%.2f)" % (n, m, sd)
  print "-" * 70
  entries = ["tech metric mean sd n"]
  for technique,results in final_report.iteritems():
    for metric,valueList in sorted(results.iteritems()):
      m, sd = mathlib.msd(valueList)
      entry = " ".join(str(x) for x in [technique, metric, m, sd, n])
      entries.append(entry)
  table = "\n".join(entries)
  writeFile("lossless.res", table)


if __name__ == '__main__':
  if len(sys.argv) > 1 and sys.argv[1] == "?":
    sys.exit('Usage: ls ./logs-bpa/*.log | python %s' % sys.argv[0])
  else:
    sys.exit(main())
