"""
This script applies lossless compression over a mouse cursor dataset.
A `lossy+lossless.res' file will be created on the script dir.
"""

# Usage example (LAB data):  find ./logs-feild/ -name "toolbar_log.txt" -print | python eval-lossy+lossless.py
# Usage example (LIVE data): ls ./logs-bpa/*.log | python eval-lossy+lossless.py

from __future__ import division
import sys, math, os, time
from core import *
from scripts import mathlib
from scripts.utils import *
from scripts.resampler import Resampler
from scripts.idt import IDT
from scripts.polling import compressByTimePolling, compressByPausedPositions
from collections import defaultdict
from scripts.delta import deltaCompression
from scripts.huffman import Huffman
from scripts.lzw import LZW
import zlib
import lz4 # Get it at http://pypi.python.org/pypi/lz4


result = {}

def applyLosslessCompression(coords, lossy_technq, lossy_compress_time, lossy_siz_comp, lossy_siz_orig, lossy_len_comp, lossy_len_orig):
  """Applies lossless compression over lossy-compressed data."""
  coordstr = str(coords)
  siz_orig = len(coordstr)
  # Method 1: Huffman
  huf = Huffman()
  start_time = time.time()
  compressed = huf.encode(coordstr)
  compress_time = time.time() - start_time
  siz_comp = (len(compressed)+7)/8 + len(str(huf.tree))
  fillResults(result, lossy_technq + '+Huffman', compress_time, siz_comp, siz_orig, 
    lossy_compress_time, lossy_siz_comp, lossy_siz_orig, lossy_len_comp, lossy_len_orig)
  # Method 2: LZW
  lzw = LZW()
  start_time = time.time()
  compressed = lzw.encode(coordstr)
  compress_time = time.time() - start_time
  siz_comp = len(compressed)
  fillResults(result, lossy_technq + '+LZW', compress_time, siz_comp, siz_orig, 
    lossy_compress_time, lossy_siz_comp, lossy_siz_orig, lossy_len_comp, lossy_len_orig)
  # Method 3: gzip
  start_time = time.time()
  compressed = coordstr.encode('zlib')
  compress_time = time.time() - start_time
  siz_comp = len(compressed)  
  fillResults(result, lossy_technq + '+Gzip', compress_time, siz_comp, siz_orig, 
    lossy_compress_time, lossy_siz_comp, lossy_siz_orig, lossy_len_comp, lossy_len_orig)
  # Method 4: LZ4
  start_time = time.time()
  compressed = lz4.dumps(coordstr)
  compress_time = time.time() - start_time
  original = lz4.loads(compressed)
  fillResults(result, lossy_technq + '+LZ4', compress_time, siz_comp, siz_orig, 
    lossy_compress_time, lossy_siz_comp, lossy_siz_orig, lossy_len_comp, lossy_len_orig)
  # Method 5: Deltas
  start_time = time.time()
  compressed = deltaCompression(coords)
  compress_time = time.time() - start_time
  siz_comp = len(str(compressed))
  fillResults(result, lossy_technq + '+Delta', compress_time, siz_comp, siz_orig, 
    lossy_compress_time, lossy_siz_comp, lossy_siz_orig, lossy_len_comp, lossy_len_orig)  
  

def processCursor(log, percentage_points):
  compression_level = 1.0 - percentage_points
  track = log.format(log.grep('move'))
  
  coords = map(lambda (x,y): [x,y], [tup[1:3] for tup in track])  
  len_orig = len(coords)
  siz_orig = len(str(coords))
  np = int(math.ceil(percentage_points * len_orig))
  
  originalPoints = reformatPoints(track)  
  siz_orig = len(str(originalPoints))

  # Method 1: WKM
  # Discarded because it's a hill climbing algorithm and therefore needs some iterations to reach local minima.
  
  # Method 2: IDT.
  dist, Ln = mathlib.distlist(coords)
  sortdist = sorted(dist, reverse=True)
  c = IDT(coords, sortdist[np - 2])
  start_time = time.time()
  c.cluster()
  compress_time = time.time() - start_time
  compressedPoints = cleanCoordsByIndex(track, c.getIndices())
  compressedPoints = assignTimestamps(c.centroids, [track[i][3] for i in c.boundaries])
  len_comp = len(compressedPoints)
  siz_comp = len(str(compressedPoints))
  applyLosslessCompression(compressedPoints, 'IDT', compress_time, siz_comp, siz_orig, len_comp, len_orig)
  
  # Method 3: Linear resampling.
  c = Resampler(coords, np)
  start_time = time.time()
  c.linear()
  compress_time = time.time() - start_time
  compressedPoints = cleanCoordsByIndex(track, c.getIndices())
  compressedPoints = assignTimestamps(c.centroids, [track[i][3] for i in c.boundaries])
  len_comp = len(compressedPoints)
  siz_comp = len(str(compressedPoints))
  applyLosslessCompression(compressedPoints, 'RSL', compress_time, siz_comp, siz_orig, len_comp, len_orig)
  
  # Method 4: Non-linear resampling.
  c = Resampler(coords, np)
  start_time = time.time()
  c.nonlinear()
  compress_time = time.time() - start_time
  compressedPoints = cleanCoordsByIndex(track, c.getIndices())
  compressedPoints = assignTimestamps(c.centroids, [track[i][3] for i in c.boundaries])
  len_comp = len(compressedPoints)
  siz_comp = len(str(compressedPoints))
  applyLosslessCompression(compressedPoints, 'RSN', compress_time, siz_comp, siz_orig, len_comp, len_orig)
  
  # Method 5: make pollInterval dependent on number of points.
  ptime = log.sesstime()*1000/np
  start_time = time.time()
  compressedPoints = compressByTimePolling(originalPoints, ptime)
  compress_time = time.time() - start_time
  len_comp = len(compressedPoints)
  siz_comp = len(str(compressedPoints))
  applyLosslessCompression(compressedPoints, 'TBP', compress_time, siz_comp, siz_orig, len_comp, len_orig)

  # Method 6: make pauseThreshold dependent on number of points.
  pauseThreshold = 20
  seen_pauses = defaultdict(int)
  while True:
    start_time = time.time()
    compressedPoints = compressByPausedPositions(originalPoints, pauseThreshold)
    compress_time = time.time() - start_time
    nc = len(compressedPoints)
    if seen_pauses[pauseThreshold] > 1:
      break
    elif nc < np:
      pauseThreshold -= 10
    elif nc > np: 
      pauseThreshold += 10
    seen_pauses[pauseThreshold] += 1
  len_comp = len(compressedPoints)
  siz_comp = len(str(compressedPoints))
  applyLosslessCompression(compressedPoints, 'PBP', compress_time, siz_comp, siz_orig, len_comp, len_orig)

  # That's all folks!
  return result


def fillResults(result, technique, 
                lossless_compress_time, lossless_siz_comp, lossless_siz_orig, 
                lossy_compress_time, lossy_siz_comp, lossy_siz_orig, lossy_len_comp, lossy_len_orig):
  lossless_compress_time *= 1000 # ms, lower is better
  lossy_compress_time *= 1000
  result[technique] = {
     # overall compression time, after the second pass
    "timesComp": lossy_compress_time + lossless_compress_time,
     # time diff, after the first pass
    "timesCompDiff": lossless_compress_time,
     # overall compression size, after the second pass
    "bytesComp": lossless_siz_comp,
    # compression diff, after the second pass
    "bytesCompDiff": lossy_siz_comp - lossless_siz_comp, 
    # size after after the first pass
    "bytesOrig": lossy_siz_orig, 
    # size diff after the first pass
    "bytesOrigDiff": lossy_siz_orig - lossless_siz_orig, 
    # these are for reference only, since we are always compressing at 50% level
    "numPtComp": lossy_len_comp,
    "numPtOrig": lossy_len_orig
  }


def reportStruct(pt_percent):
  report = {}
  values = [
              "timesComp", "bytesComp", "bytesOrig", "numPtComp", "numPtOrig", 
              "timesCompDiff", "bytesCompDiff", "bytesOrigDiff"
           ]
  technq = [
              "IDT+Huffman", "IDT+LZW", "IDT+Gzip", "IDT+LZ4", "IDT+Delta",
              "RSL+Huffman", "RSL+LZW", "RSL+Gzip", "RSL+LZ4", "RSL+Delta",
              "RSN+Huffman", "RSN+LZW", "RSN+Gzip", "RSN+LZ4", "RSN+Delta",
              "TBP+Huffman", "TBP+LZW", "TBP+Gzip", "TBP+LZ4", "TBP+Delta",
              "PBP+Huffman", "PBP+LZW", "PBP+Gzip", "PBP+LZ4", "PBP+Delta"
           ]
  for t in technq:
    if t not in report:
      report[t] = {}
    for v in values:
      if v not in report[t]:
        report[t][v] = {}
      if pt_percent not in report[t][v]:
        report[t][v][str(pt_percent)] = []
  return report


def main():
  pt_percent   = 50
  final_report = reportStruct(pt_percent)
  for num,log in enumerate(getLogs()): # reads from stdin
    res = processCursor(log, pt_percent/100.0)
    for technique,results in res.iteritems():
      for metric,value in sorted(results.iteritems()):
        final_report[technique][metric][str(pt_percent)].append(value)
  if num > 0 and num % 1000 == 0:
    print num,"logs processed"
  n = num + 1 # We need this to build CIs offline
  entries = ["tech metric perc nones mean sd n"]
  for technique,results in final_report.iteritems():
    for metric,percentages in sorted(results.iteritems()):
      for perc,valueList in sorted(percentages.iteritems()):
        cleanList = removeNones(valueList)
        if cleanList:
          m, sd = mathlib.msd(cleanList)
          entry = " ".join(str(x) for x in [technique, metric, perc, countNones(valueList), m, sd, n])
          entries.append(entry)
  table = "\n".join(entries)
  writeFile("lossy+lossless.res", table)


if __name__ == '__main__':
  if len(sys.argv) > 1 and sys.argv[1] == "?":
    sys.exit('Usage: ls ./logs-bpa/*.log | python %s' % sys.argv[0])
  sys.exit(main())
