"""
This script applies lossless compression over a mouse cursor dataset.
A `lossy.res' file will be created on the script dir.
"""

# Usage example (LAB data):  find ./logs-feild/ -name "toolbar_log.txt" -print | python eval-lossy.py
# Usage example (LIVE data): ls ./logs-bpa/*.log | python eval-lossy.py

from __future__ import division
import sys, math, os, time
from core import *
from scripts import mathlib
from scripts.utils import *
from scripts.resampler import Resampler
from scripts.idt import IDT
from scripts.polling import compressByTimePolling, compressByPausedPositions
from collections import defaultdict


def processCursor(log, percentage_points):
  compression_level = 1.0 - percentage_points
  result = {}
  track  = log.format(log.grep('move'))
  
  coords = map(lambda (x,y): [x,y], [tup[1:3] for tup in track])  
  len_orig = len(coords)
  siz_orig = len(str(coords))
  np = int(math.ceil(percentage_points * len_orig))
  
  originalPoints = reformatPoints(track)
  siz_orig = len(str(originalPoints))

  clicks = log.format(log.grep('click'))
  clicktimes = [tup[3] for tup in clicks]

  # Method 1: WKM
  # Discarded because it's a hill climbing algorithm and therefore needs some iterations to reach local minima.
  
  # Method 2: IDT
  dist, Ln = mathlib.distlist(coords)
  sortdist = sorted(dist, reverse=True)
  c = IDT(coords, sortdist[np - 2])
  start_time = time.time()
  c.cluster()
  compress_time = time.time() - start_time
  compressedPoints = cleanCoordsByIndex(track, c.getIndices())
  compressedPoints = assignTimestamps(c.centroids, [track[i][3] for i in c.boundaries])
  len_comp = len(compressedPoints)
  siz_comp = len(str(compressedPoints))
  dd, dc = evalCompression(originalPoints, compressedPoints, clicktimes)
  fillResults(result, 'IDT', compress_time, siz_comp, siz_orig, len_comp, len_orig, dd, dc)
  
  # Method 3: Linear resampling.
  c = Resampler(coords, np)
  start_time = time.time()
  c.linear()
  compress_time = time.time() - start_time
  compressedPoints = cleanCoordsByIndex(track, c.getIndices())
  compressedPoints = assignTimestamps(c.centroids, [track[i][3] for i in c.boundaries])
  len_comp = len(compressedPoints)
  siz_comp = len(str(compressedPoints))
  dd, dc = evalCompression(originalPoints, compressedPoints, clicktimes)
  fillResults(result, 'RSL', compress_time, siz_comp, siz_orig, len_comp, len_orig, dd, dc)
  
  # Method 4: Non-linear resampling.
  c = Resampler(coords, np)
  start_time = time.time()
  c.nonlinear()
  compress_time = time.time() - start_time
  compressedPoints = cleanCoordsByIndex(track, c.getIndices())
  compressedPoints = assignTimestamps(c.centroids, [track[i][3] for i in c.boundaries])
  len_comp = len(compressedPoints)
  siz_comp = len(str(compressedPoints))
  dd, dc = evalCompression(originalPoints, compressedPoints, clicktimes)
  fillResults(result, 'RSN', compress_time, siz_comp, siz_orig, len_comp, len_orig, dd, dc)
  
  # Method 5: make pollInterval dependent on number of points.
  ptime = log.sesstime()*1000/np
  start_time = time.time()
  compressedPoints = compressByTimePolling(originalPoints, ptime)
  compress_time = time.time() - start_time
  len_comp = len(compressedPoints)
  siz_comp = len(str(compressedPoints))
  dd, dc = evalCompression(originalPoints, compressedPoints, clicktimes)
  fillResults(result, 'TBP', compress_time, siz_comp, siz_orig, len_comp, len_orig, dd, dc)

  # Method 6: make pauseThreshold dependent on number of points.
  pauseThreshold = 20
  seen_pauses = defaultdict(int)
  while True:
    start_time = time.time()
    compressedPoints = compressByPausedPositions(originalPoints, pauseThreshold)
    compress_time = time.time() - start_time
    nc = len(compressedPoints)
    if seen_pauses[pauseThreshold] > 1:
      break
    elif nc < np:
      pauseThreshold -= 10
    elif nc > np: 
      pauseThreshold += 10
    seen_pauses[pauseThreshold] += 1
  len_comp = len(compressedPoints)
  siz_comp = len(str(compressedPoints))
  dd, dc = evalCompression(originalPoints, compressedPoints, clicktimes)
  fillResults(result, 'PBP', compress_time, siz_comp, siz_orig, len_comp, len_orig, dd, dc)
  
  # That's all folks!
  return result

def fillResults(result, technique, compress_time, siz_comp, siz_orig, len_comp, len_orig, dd, dc):
  result[technique] = {
    "timesComp": compress_time * 1000,          # ms, less is better
    "compRatio": (1 - siz_comp/siz_orig) * 100, # %, higher is better
    "numPtComp": len_comp,                      # num. compressed points
    "numPtOrig": len_orig,                      # num. original points
    "bytesComp": siz_comp,                      # compressed size
    "bytesOrig": siz_orig,                      # original size
    "evalDista": dd,                            # trail replication (px)
    "evalClick": dc                             # trail replication at clicks (px)
  }
  
def reportStruct(pt_percent):
  report = {}
  values = ["timesComp", "compRatio", "numPtComp", "numPtOrig", "bytesComp", "bytesOrig", "evalDista", "evalClick"]
  technq = ["IDT", "RSL", "RSN", "TBP", "PBP"]
  for t in technq:
    if t not in report:
      report[t] = {}
    for v in values:
      if v not in report[t]:
        report[t][v] = {}
      for c in pt_percent:
        if c not in report[t][v]:
          report[t][v][str(c)] = []
  return report

def main():
  pt_percent   = sorted([p for p in range(10, 100, 10)], reverse=True)
  final_report = reportStruct(pt_percent)
  browse_times = []
  for num,log in enumerate(getLogs()): # reads from stdin
    for perc in pt_percent:
      res = processCursor(log, perc/100.0)
      for technique,results in res.iteritems():
        for metric,value in sorted(results.iteritems()):
          final_report[technique][metric][str(perc)].append(value)
    browse_times.append(log.sesstime()/60) # minutes
    if num > 0 and num % 1000 == 0:
      print num,"logs processed"
  print
  m, sd = mathlib.msd(browse_times)
  n = len(browse_times) # We need this to build CIs offline
  print "%d users browsed for %f minutes (SD=%.2f)" % (n, m, sd)
  print "-" * 70
  entries = ["tech metric perc nones mean sd n"]
  for technique,results in final_report.iteritems():
    for metric,percentages in sorted(results.iteritems()):
      for perc,valueList in sorted(percentages.iteritems()):
        cleanList = removeNones(valueList)
        if cleanList:
          m, sd = mathlib.msd(cleanList)
          entry = " ".join(str(x) for x in [technique, metric, perc, countNones(valueList), m, sd, n])
          entries.append(entry)
  table = "\n".join(entries)
  writeFile("lossy.res", table)

if __name__ == '__main__':
  if len(sys.argv) > 1 and sys.argv[1] == "?":
    sys.exit('Usage example: ls ./logs-bpa/*.log | python %s' % sys.argv[0])
  sys.exit(main())
