"""
This script tests all compression algorithms evaluated in our paper:
Luis A. Leiva and Jeff Huang. Building a Better Mousetrap: 
Compressing Mouse Cursor Activity for Web Analytics. 
Information Processing & Management. Elsevier.
"""

from __future__ import  division
import sys, math, os
import mathlib
from parser import EvLogParser, ToolbarLogParser
from resampler import Resampler
from delta import deltaCompression
from idt import IDT
from polling import *
from huffman import Huffman
from lzw import LZW
import zlib
import lz4 # Get it at http://pypi.python.org/pypi/lz4


def usage():
  print "Usage: python %s <log_file> <compression_level in [10,...,90]>" % sys.argv[0]

def main(filename, compression):
  basename, extension = os.path.basename(filename).split('.')
  if extension == 'xml':
    log = EvXMLParser(filename)
    processXML(filename)
  elif extension == 'log':
    log = EvLogParser(filename)
    processCursor(log, compression)
  elif extension == 'txt':
    log = ToolbarLogParser(filename)
    processCursor(log, compression)
  else:
    sys.exit("Error: Unknown log extension")

def assignTimestamps(points, timestamps):
  assert len(points) == len(timestamps)
  return [(timestamps[i], pt[0], pt[1]) for i, pt in enumerate(points)]

def reformatPoints(points):
  return map(lambda (x,y,t): (t,x,y), [tup[1:4] for tup in points])
    
def printpts(c, g):
  return len(c)/len(g)

def processCursor(log, compression):
  separator = "-"*70
  print separator
  print "Loading %s. Browsing time is %.2f min. Compression set to %.2f" % (log.fname, log.sesstime()/60, compression)
  print separator  
  rows = log.grep('move') # retrieve only (mouse|touch)move events
  track = log.format(rows)
  
  # LOSSY COMPRESSION ----------------------------------------------------------
  
  # Resampling algorithms need arrays of coordinates, so let's convert log track tuples.
  coords = map(lambda (x,y): [x,y], [tup[1:3] for tup in track])
  k = len(coords) * compression/100 # number of compressed points based on compresion ratio.

  # Method 1: IDT
  dist, Ln = mathlib.distlist(coords)
  # We can choose dt ~ mean(dist) + 2*sd(dist), kind of 95% CI.
  # However, we want dt such that IDT returns a desired number of points.
  sortdist = sorted(dist, reverse=True)
  numpts = int(k) - 2 # substract 2 because IDT uses distance thresholds and the first and last points are added in getIndices().
  c = IDT(coords, sortdist[numpts])
  c.cluster()
  compressedPoints = assignTimestamps(c.centroids, [track[i][3] for i in c.boundaries])
  print "Testing IDT:", printpts(compressedPoints, coords)
  
  # Now try basic resampling methods (linear, non-linear, plus in/ex versions).
  c = Resampler(coords, k)
  
  # Method 2.a) resampled points belong to the original dataset.
  c.linear() 
  compressedPoints = assignTimestamps(c.centroids, [track[i][3] for i in c.boundaries])
  print "Testing linear (in) resampling:", printpts(compressedPoints, coords)
  
  # Method 2.b) resampled points don't belong to the original dataset.
  c.linear(False) 
  compressedPoints = assignTimestamps(c.centroids, [track[i][3] for i in c.boundaries])
  print "Testing linear (out) resampling:", printpts(compressedPoints, coords)
  
  # Method 2.c) same as 2.a
  c.nonlinear() 
  compressedPoints = assignTimestamps(c.centroids, [track[i][3] for i in c.boundaries])
  print "Testing nonlinear (in) resampling:", printpts(compressedPoints, coords)
  
  # Method 2.d) same as 2.b
  c.nonlinear(False) 
  compressedPoints = assignTimestamps(c.centroids, [track[i][3] for i in c.boundaries])
  print "Testing nonlinear (out) resampling:", printpts(compressedPoints, coords)

  # Now try polling techniques (time-based, pause-based).
  originalPoints = reformatPoints(track)
  
  # Method 3.a) make pollInterval dependent on number of points.
  ptime = log.sesstime()*1000/k
  compressedPoints = compressByTimePolling(originalPoints, ptime)
  print "Testing time-based polling:", printpts(compressedPoints, originalPoints)
    
  # Method 3.b) make pollInterval dependent on number of points.
  compressedPoints = compressByPausedPositions(originalPoints)
  print "Testing pause-based polling:", printpts(compressedPoints, originalPoints)


  # LOSSLESS COMPRESSION -------------------------------------------------------
  
  coords = map(lambda (x,y,t): [x,y,t], [tup[1:4] for tup in track])
  coordstr = str(coords)
  siz_orig = len(coordstr)
  
  # Method 1: Huffman
  huf = Huffman()
  compressed = huf.encode(coordstr)
  original = huf.decode(compressed)
  assert coordstr == original
  siz_comp = (len(compressed) + 7)/8 + len(str(huf.tree)) # convert bits to bytes, plus enconding tree
  print "Testing Huffman encoding:", siz_comp/siz_orig
  
  # Method 2: LZW
  lzw = LZW()
  compressed = lzw.encode(coordstr)
  original = lzw.decode(compressed)
  assert coordstr == original
  siz_comp = len(compressed)
  print "Testing LZW encoding:", siz_comp/siz_orig
    
  # Method 3: gzip
  compressed = coordstr.encode('zlib')
  original = compressed.decode('zlib')
  assert coordstr == original
  siz_comp = len(compressed)  
  print "Testing Gzip encoding:", siz_comp/siz_orig

  # Method 4: LZ4
  compressed = lz4.dumps(coordstr)
  original = lz4.loads(compressed)
  assert coordstr == original
  siz_comp = len(compressed)
  print "Testing LZ4 encoding:", siz_comp/siz_orig

  # Method 5: Delta
  compressed = deltaCompression(coords)
  siz_comp = len(str(compressed))
  print "Testing Delta encoding:", siz_comp/siz_orig
  
  
if __name__ == '__main__':
  if len(sys.argv[1:]) < 2:
    sys.exit(usage())
  compression = float(sys.argv[2])
  if compression < 10 or compression > 90:
    print "Error: compression level out of bounds"
    sys.exit(usage())
  else:
    sys.exit(main(sys.argv[1], compression))
