import math
from mathlib import sqL2

class IDT:
  """
  Dispersion-Threshold Identification (I-DT) algorithm, by Luis A. Leiva.
  See https://www.cs.drexel.edu/~salvucci/publications/Salvucci-ETRA00.pdf
  Dual licensed under the MIT and GPL licenses.
  """
  
  def __init__(self, samples,threshold=0.0):
    """
    Instantiates the class.
    The threshold arg is the distance threshold that will be used to select candidate points.
    """
    # User input
    self.samples    = samples
    self.threshold  = float(threshold)
    # Assume that all samples have the same dimensions
    self.dimensions = len(self.samples[0])
    # Sanitize user input
    if threshold < 0:
      self.threshold = 0.0
    self.reset()


  def reset(self):
    """Clean up."""
    # System ouput
    self.centroids  = []
    self.boundaries = []


  def cluster(self):
    """Compress samples."""
    self.reset()
    N = len(self.samples)
    for i in range(1,N):
      d = sqL2(self.samples[i-1], self.samples[i])
      if math.sqrt(d) > self.threshold:
        self.boundaries.append(i)
    # IDT may fail to retrieve the correct segmentation if threshold is too high
    if not len(self.boundaries):
      self.boundaries = [0, N-1]
    self.centroids = [self.samples[s] for s in self.boundaries]


  def getIndices(self):
    """Get data point indices."""
    assert len(self.boundaries) > 0
    ind = self.boundaries[:]
    firstIndex = 0
    lastIndex = len(self.samples) - 1
    if ind[0] != firstIndex:
      ind.insert(0, firstIndex)
    if ind[-1] != lastIndex:
      ind.append(lastIndex)
    return ind
