"""
Math utilities for vector data, by Luis A. Leiva.
Dual licensed under the MIT and GPL licenses.
"""

import math
from operator import mul

def msd(vec):
  """Computes the Mean and SD of vector [v1,...,vN]."""
  N, sd = len(vec), 0.0
  mean = sum(vec)/float(N)
  if len(vec) > 1:
    for v in vec:
      sd += (v - mean)**2
    sd = math.sqrt(sd / (N-1))
  return mean, sd

def distlist(samples):
  """Computes distances and length of a vector sequence [ [v11,...,v1N], ... [vn1,...,vnN] ]."""
  N, l, Ln = len(samples), [], 0.0
  for i in range(1,N):
    Li = math.sqrt( sqL2(samples[i], samples[i-1]) )
    Ln += Li
    l.append(Li)
  return l, Ln

def sqL2(a,b):
  """Computes the squared L2 distance between 2 vectors [a1,...,aN] and [b1,...,bN]."""
  dim, nrg = len(a), 0.0
  for d in range(dim):
    dist = a[d] - b[d]
    nrg += dist * dist
  return nrg

def clusterCenter(samples):
  """Computes the center of a vector sequence [ [v11,...,v1N], ... [vn1,...,vnN] ]."""
  N, dim = len(samples), len(samples[0])
  if N == 1: # singleton cluster
    return samples[0]
  dsum = [0.0] * dim
  for d in range(dim):
    for i in range(N):
      dsum[d] += samples[i][d]
    dsum[d] /= N
  return dsum

def binarySearch(a, x, lo=0, hi=None):
  """Performs binary search over vector [a1,...,aN]."""
  if hi is None:
    hi = len(a)
  while lo < hi:
    mid = (lo + hi)//2
    midval = a[mid]
    if midval < x:
      lo = mid + 1
    elif midval > x: 
      hi = mid
    else:
      return mid
  return -1
