import xml.dom.minidom as dom
from xml.parsers.expat import ExpatError
import re, htmlentitydefs
import string
from utils import *

class LogParser:
  """
  A general parser for mouse cursor logs, by Luis A. Leiva.
  This abstract class unifies both log parsers (Feild's toolbar logs and best paper awards logs).
  """
    
  def filter(self, evtype):
    """Allows filtering log rows by event type."""
    evtype = evtype.lower()
    return [t for t in self.track if evtype == t[1].lower()]

  def filterList(self, evlist):
    """Allows filtering log rows by a list of event types."""
    evlist = map(string.lower, evlist)
    return [t for t in self.track if t[1].lower() in evlist]

  def grep(self, evpattern):
    evpattern = evpattern.lower()
    """Allows filtering log rows by matching a string in event type."""
    return [t for t in self.track if evpattern in t[1].lower()]
    
  def sesstime(self):
    """Computes session time in seconds."""
    return (self.track[-1][4] - self.track[0][4]) / 1000.0

  def getCursors(self, rows=None):
    """Retrieves a cursor dictionary from given rows or internal mouse cursor track."""
    cursors = {}
    l = self.track if rows is None else rows
    for id, event, x, y, time in l:
      if id not in cursors:
        cursors[id] = []
      tup = (event, x, y, time)
      cursors[id].append(tup)
    return cursors

  def format(self, rows=None):
    """Formats given rows or internal mouse cursor track."""
    l = self.track if rows is None else rows
    return [ (event, x, y, time) for id, event, x, y, time in l ]


class EvLogParser(LogParser):
  """
  A parser for evtrack logs, by Luis A. Leiva.
  Each row is cursor_id, timestamp, x,y, event_name, related_element.
  """
  
  def __init__(self, filename, allownc=False):
    """
    Instantiates the class. 
    The flag 'allownc' controls whether null (0,0) coordinates should be included.
    """
    self.fname = filename
    self.track = []
    lines = open(filename).readlines()
    for line in uniq(lines):
      lineparts = line.split(" ")
      if len(lineparts) < 6:
        continue
      id, time, x,y, event, element = lineparts
      if (int(x) > 0 and int(y) > 0) or allownc:
        tup = (id, event, int(x), int(y), int(time))
        self.track.append(tup)


class ToolbarLogParser(LogParser):
  """
  A parser for Feild's toolbar logs, by Luis A. Leiva.
  """
  
  def __init__(self, filename, allownc=False):
    """
    Instantiates the class. 
    The flag 'allownc' controls whether null (0,0) coordinates should be included.
    """
    self.fname = filename
    self.track = []
    lines = open(filename).readlines()
    for num, line in enumerate(lines):
      row = line.split("\t")
      num_entries, row_event, row_time = len(row), row[0], row[1]
      # Workaround to retrieve "correct" click info from Feild's data.
      if num_entries < 11 and "click" not in row_event.lower(): 
        continue
      if num_entries == 11:
        event, time, tab_id, x, y, iw, ih, sx, sy, ow, oh = row
      elif "click" in row_event.lower():
        id, event, x, y, time = self.track[-1]
        event, time = row_event, row_time
      if (int(x) > 0 and int(y) > 0) or allownc:
        # Set cursor ID to 0, since only 1 cursor was considered in Feild's data.
        tup = (0, event, int(x), int(y), int(time))
        self.track.append(tup)


class EvXMLParser:
  """
  A parser for evtrack XMLs, by Luis A. Leiva.
  """
  
  def __init__(self, filename):
    """Instantiates the class."""  
    self.fname = filename
    try:
      self.doc = dom.parse(filename)
    except ExpatError:
      xmlstr = open(filename).read().replace("&", "&amp;")
      self.doc = dom.parseString(xmlstr)

  def get(self, name):
    """Main interface. Retrieves the value of a given node name."""
    return self.getText(self.getNode(name))
    
  def getNode(self, nodeName):
    return self.doc.getElementsByTagName(nodeName)[0].childNodes
    
  def getText(self, nodelist):
    rc = []
    for node in nodelist:
      if node.nodeType == node.TEXT_NODE:
        rc.append(node.data)
    return ''.join(rc)
