import math
from mathlib import distlist, sqL2, clusterCenter

class Resampler:
  """
  A collection of resampling methods for vector data, by Luis A. Leiva.
  Dual licensed under the MIT and GPL licenses.
  """
  
  def __init__(self, samples, numclusters=1):
    """
    Instantiates the class.
    The numclusters arg indicates how many points are going to be selected.
    """
    # User input
    self.samples     = samples
    self.numclusters = int(numclusters)
    # Assume that all samples have the same dimensions
    self.dimensions = len(self.samples[0])
    # Sanitize user input
    if numclusters <= 0:
      self.numclusters = 1
    elif numclusters > len(samples):
      self.numclusters = len(samples)


  def reset(self):
    """Clean up."""
    # System ouput
    self.centroids  = []
    self.boundaries = []


  def linear(self, preserve=True):
    """
    Performs a linear resampling of N points to M points.
    The preserve flag indicates whether interpolated points must belong to the original dataset or not.
    """
    self.reset()
    N, M = len(self.samples), self.numclusters
    if M < 2:
      self.centroids  = [clusterCenter(self.samples)] # single partition
      self.boundaries = [0]
      return
    elif M >= N:
      self.centroids  = self.samples # singleton clusters
      self.boundaries = range(N)
      return    
    if preserve:
      self.linearIn(N,M)
    else:
      self.linearEx(N,M)

    
  def linearIn(self, N, M):
    """Linear resampling where each point comes from the original dataset."""  
    self.reset()
    b = -1
    for i in range(0,N):
      q = math.floor( (i+1)*M/(N+1.0) )
      if q > b:
        b = q
        self.centroids.append(self.samples[i])
        self.boundaries.append(i)


  def linearEx(self, N, M):
    """Linear resampling where each point does not belong to the original dataset."""
    self.reset()
    Lcum, LN = distlist(self.samples)
    incr, D = LN / float(M-1), 0.0
    self.centroids.append(self.samples[0])
    self.boundaries.append(0)
    for i in range(1,N):
      dist = Lcum[i] - Lcum[i-1]
      if (D + dist) >= incr:
        yp = range(self.dimensions)
        for d in range(0,self.dimensions):
          yp[d] = self.samples[i-1][d] + (self.samples[i][d] - self.samples[i-1][d]) * (incr - D) / dist
        self.centroids.append(yp)
        self.boundaries.append(i)
        D = 0.0
      else: 
        D += dist
    self.centroids.append(self.samples[N-1])
    self.boundaries.append(N-1)


  def nonlinear(self, preserve=True):
    """
    Performs a non-linear resampling of N points to M points (TS algorithm).
    The preserve flag indicates whether interpolated points must belong to the original dataset or not.
    """
    self.reset()
    N, M = len(self.samples), self.numclusters
    if M < 2:
      self.centroids  = [clusterCenter(self.samples)] # single partition
      self.boundaries = [0]
      return
    elif M >= N:
      self.centroids  = self.samples # singleton clusters 
      self.boundaries = range(N)
      return
    if preserve:
      self.nonlinearIn(N,M)
    else:
      self.nonlinearEx(N,M)


  def nonlinearIn(self, N, M):
    """Non-linear resampling where each point comes from the original dataset."""
    self.reset()
    Lcum, Ln = distlist(self.samples)
    incr, i = Ln / float(M), 1
    for j in range(1,M+1):
      fact = (j - 1)*incr
      while fact > Lcum[i]:
        i += 1
      self.centroids.append(self.samples[i-1])
      self.boundaries.append(i-1)


  def nonlinearEx(self, N, M):
    """Non-linear resampling where each point does not belong to the original dataset."""
    self.reset()
    Lcum, Ln = distlist(self.samples)
    incr, i = Ln / float(M), 1
    for j in range(1,M+1):
      fact = (j - 1)*incr
      while fact > Lcum[i]:
        i += 1
      yp = [0.0] * self.dimensions
      for d in range(self.dimensions):
        yp[d] = self.samples[i-1][d] + (self.samples[i][d] - self.samples[i-1][d]) * (fact - Lcum[i-1]) / (Lcum[i] - Lcum[i-1])
      self.centroids.append(yp)
      self.boundaries.append(i)


  def getIndices(self):
    """Return data point indices of the original dataset."""
    assert len(self.boundaries) > 0
    ind = self.boundaries[:]
    firstIndex = 0
    lastIndex = len(self.samples) - 1
    if ind[0] != firstIndex:
      ind.insert(0, firstIndex)
    if ind[-1] != lastIndex:
      ind.append(lastIndex)
    return ind
